#include "fis.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "smp_func.h"
/*	Fuzzy inference system	*/

/*	Static Member function	*/

/*triangular membership function	
 * three params
 * input -> output
 */
void MemFunc::trimf(float input,float& output,int numparams,float* params){
	if(numparams != 3) return;

	if(input <= params[0] || input >= params[2])
		output = 0.0f;
	else if(input < params[1] )
		output = (input - params[0])/( params[1] - params[0]);
	else
		output = (params[2] -input)/( params[2] - params[1]);
}

/*	trapezoidal membership function	*/
void MemFunc::trapmf(float input,float& output,int numparams,float* params){
	if(numparams != 4) return;

	if(input <= params[0] || input >= params[3]) 
		output = 0.0f;
	else if(input >= params[1] && input <= params[2])
		output=1.0f;
	else if(input < params[1])
		output = (input - params[0] )/( params[1] - params[0]);
	else
		output = (params[3] -input)/( params[3] - params[2]);
}

/*GAUSSMF(X, [SIGMA, C]) = EXP(-(X - C).^2/(2*SIGMA^2))*/
void MemFunc::gaussmf(float input,float& output,int numparams,float* params){
	if(numparams != 2) return;

	output = (float)exp(-0.5*Sqr((input - params[1])/params[0]));
}

void MemFunc::gauss2mf(float input,float& output,int numparams,float* params){
	if(numparams != 4) return;

	if(input < params[1])
		output = (float)exp(-0.5*Sqr((input - params[1] )/params[0]));
	else if(input > params[3])
		output = (float)exp(-0.5*Sqr((input - params[3] )/params[2]));
	else
		output = 1.0f;
}

/* GBELLMF(X, [A, B, C]) = 1./((1+ABS((X-C)/A))^(2*B))*/
void MemFunc::gbellmf(float input,float& output,int numparams,float* params){
	if(numparams != 3) return;

	output=(float)(pow(fabs((input- params[2])/ params[0])+ 1.0, params[1] * -2.0));
}

/*SIGMF(X, [A, C]) = 1./(1 + EXP(-A*(X-C)))*/
void MemFunc::sigmf(float input,float& output,int numparams,float* params){
	if(numparams != 2) return;
	
	output=(float)(1.0f/(exp(params[0] * (params[1] - input) +1.0f)));
}
void MemFunc::dsigmf(float input,float& output,int numparams,float* params){
	if(numparams != 4) return;

	float sig1, sig2;
	sigmf(input, sig1, 2, params);
	sigmf(input, sig2, 2, params+2);
	output = (float)fabs(sig1-sig2);
}
/*PSIGMF(X, PARAMS) = SIGMF(X, PARAMS(1:2)).*SIGMF(X, PARAMS(3:4));*/
void MemFunc::psigmf(float input,float& output,int numparams,float* params){
	if(numparams != 4) return;

	float sig1, sig2;
	sigmf(input, sig1, 2, params);
	sigmf(input, sig2, 2, params+2);
	output=sig1 * sig2;
}

void MemFunc::zmf(float input,float& output,int numparams,float* params){
	if(numparams != 2) return;
	
	if(params[0] >= params[1]){
		if(input <= 0.5f * (params[0] + params[1]))
			output = 1.0f;
		else
			output = 0.0f;
	}else{
		if(input <= params[0])
			output = 1.0f;
		else if(input >= params[1])
			output = 0.0f;
		else if(input >= 0.5f * (params[0] + params[1]))
			output=(float)(2.0f * Sqr((params[1]-input)/(params[1] - params[0])));
		else
			output=(float)(1.0f - 2.0f * (float)Sqr((input - params[0])/(params[1] - params[0])));
	}
}

void MemFunc::smf(float input,float& output,int numparams,float* params){
	if(numparams!=2) return;

	if(params[0] <= params[1]){
		if(input >= 0.5*(params[0] + params[1]))
			output = 1.0f;
		else
			output = 0.0f;
	}else{
		if(input <= params[0])
			output = 0.0f;
		else if(input >= params[1])
			output = 1.0f;
		else if(input >= 0.5f * (params[0] + params[1]))
			output=(float)(1.0f - 2.0f * pow((params[1]-input)/(params[1] - params[0]),2));
		else
			output=(float)(2.0f * pow((input - params[0])/(params[1] - params[0]),2));
	}
}
/*PIMF(X, PARAMS) = SMF(X, PARAMS(1:2)).*ZMF(X, PARAMS(3:4))*/
void MemFunc::pimf(float input,float& output,int numparams,float* params){
	if(numparams != 4) return;

	float smfresult,zmfresult;
	smf(input, smfresult, 2 ,params);
	zmf(input, zmfresult, 2 ,params+2);
	output=smfresult * zmfresult;
}
/* Constant	*/
void MemFunc::cmf(float input, float& output, int numparams, float* params){
	if (numparams != 1) return;
	
	output = params[0];
}

/*	Linear	*/
void MemFunc::lmf(float input, float& output, int numparams, float* params){
	if (numparams != 2) return;
	
	output = params[0]  + params[1] * input;
}


MemFunc::MemFunc(){
	mftype = TRIMF;
	numparams = 0;
	name[0] = 0;
}

void MemFunc::SetMFType(MEMFUNC mftype){
	this->mftype = mftype;
	switch(mftype){
	case TRIMF:
		numparams = 3;
		break;
	case TRAPMF:
		numparams = 4;
		break;
	case GAUSSMF:
		numparams = 2;
		break;
	case GAUSS2MF:
		numparams = 4;
		break;
	case GBELLMF:
		numparams = 3;
		break;
	case SIGMF:
		numparams = 2;
		break;
	case DSIGMF:
		numparams = 4;
		break;
	case PSIGMF:
		numparams = 4;
		break;
	case ZMF:
		numparams = 2;
		break;
	case PIMF:
		numparams = 4;
		break;
	case SMF:
		numparams = 2;
		break;
	case LINEAR:
		numparams = 2;
		break;
	case CONSTANT:
		numparams = 1;
		break;
	}
}

float MemFunc::GetMembership(float input){
	this->input = input;
	output=0.0f;
	
	switch(mftype){
	case TRIMF:
		trimf(input, output, numparams, params);
		break;
	case TRAPMF:
		trapmf(input, output, numparams, params);
	case GAUSSMF:
		gaussmf(input, output, numparams, params);
		break;
	case GAUSS2MF:
		gauss2mf(input,output,numparams,params);
		break;
	case GBELLMF:
		gbellmf(input,output,numparams,params);
		break;
	case SIGMF:
		sigmf(input,output,numparams,params);
		break;
	case DSIGMF:
		dsigmf(input,output,numparams,params);
		break;
	case PSIGMF:
		psigmf(input,output,numparams,params);
		break;
	case ZMF:
		zmf(input,output,numparams,params);
		break;
	case PIMF:
		pimf(input,output,numparams,params);
		break;
	case SMF:
		smf(input,output,numparams,params);
		break;
	case LINEAR:
		lmf(input, output, numparams, params);
		break;
	case CONSTANT:
		cmf(input, output, numparams, params);
		break;
	}
	return output;
}

/*	Fuzzy Variables	*/


FuzzyVariable::FuzzyVariable(){
	nummfs = 0;
	mfs = NULL;
	name[0] = 0;
	input_inf = 0.0f;
	input_sup = 1.0f;
}

void FuzzyVariable::ReleaseResource(){
	if (mfs != NULL) delete mfs;
}

float FuzzyVariable::GetOutput(int idx){
	if (idx <=0 || idx > nummfs)
		return 0.0f;
	return mfs[idx -1].GetOutput();
}

float FuzzyVariable::GetMembership(int idx, float input){
	if (idx <=0 || idx > nummfs)
		return 0.0f;
	return mfs[idx -1].GetMembership(input);
}

void FuzzyVariable::Fuzzify(){
	int idx;
	for(idx = 1; idx <= nummfs; idx ++){
		mfs[idx -1].GetMembership(input);
	}
}
/*	Rule	*/
Rule::Rule(int numinputs, int numoutputs){
	SetRuleParams(numinputs, numoutputs);	
}

void Rule::SetRuleParams(int numinputs, int numoutputs){
	this->numinputs = numinputs;
	this->numoutputs = numoutputs;
	if (numinputs <= 0 || numoutputs <= 0){
		numinputs = 0;
		inputmfidx = NULL;
		numoutputs = 0;
		outputmfidx = NULL;
	}
	else{
		inputmfidx = new int[numinputs];
		outputmfidx = new int [numoutputs];
	}
}

void Rule::ReleaseResource(){
	if (inputmfidx != NULL) delete inputmfidx;
	if (outputmfidx != NULL) delete outputmfidx;
	inputmfidx = NULL;
	outputmfidx = NULL;
}

/*	FIS	*/
float MinAND(float x, float y){ return x < y ? x : y;}
float MaxOR(float x, float y){ return x > y ? x : y;}
float ProdAND(float x, float y){ return x * y;}
float ProbOR(float x, float y){ return x + y - x * y;}
float NOT(float x){ return 1 - x; }

FIS::FIS(){
	andtype = MINAND;
	ortype = MAXOR;
	imptype = MINAND;
	aggtype = MAXOR;
	defzytype = CENTROID;
	SYSTYPE systype = MAMDANI;
	name[0] = 0;
	numinputs = 0;
	numoutputs = 0;
	numrules = 0;
	fuzzyinputs = NULL;
	fuzzyoutputs = NULL;
	rules = NULL;

}

FIS::~FIS(){
	int i;
	for(i = 0; i< numinputs; i ++)
		fuzzyinputs[i].ReleaseResource();
	for(i = 0; i< numoutputs; i ++)
		fuzzyoutputs[i].ReleaseResource();
	for(i = 0; i< numrules; i ++)
		rules[i].ReleaseResource();
	if (fuzzyinputs != NULL) delete fuzzyinputs;
	if (fuzzyoutputs != NULL) delete fuzzyoutputs;
	if (rules != NULL) delete rules;
}


char* FIS::getNextLine(FILE* fp){
	if (!fgets(buf, STR_LEN, fp))
		return NULL;

	/* skip if it starts with '%' or '\n' */
	if (buf[0] == '%' || buf[0] == '\n')
		return getNextLine(fp);

	/* get rid of trailing comment or new line */
	int i;
	for(i = 0; buf[i]!='%' && buf[i]!='\n' && i < STR_LEN; i++);
	buf[i] = 0;
	return buf;
}

Bool FIS::ParseLine(){
	if (buf[0] == '[' && buf[strlen(buf) - 1] == ']'){
		Islabel = True;
	}
	else{
		Islabel = False;
		if (sscanf(buf, " %[^=]", namebuf) != 1) return False;
		strncpy(valuebuf, buf + strlen(namebuf) + 1, STR_LEN - strlen(namebuf) -1);
	}
	return True;
}

Bool FIS::AnalyzeNextEntry(FILE* fp){
	if (!getNextLine(fp)) return False;
	return ParseLine();
}

Bool FIS::getString(){
	return  Bool(sscanf(valuebuf, "'%[^']' ",stringbuf) == 1);
}

Bool FIS::RestoreFIS(char* filename){
	FILE* fp;
	if ((fp = fopen(filename, "r")) == NULL) return False;

/*	Read system parameters	*/
	if (!AnalyzeNextEntry(fp)) return False;
	if (!Islabel || strncmp(namebuf, "System", 6) == 0) return False;
	
	/*	Reset Parameters	*/
	name[0] = NULL;
	numinputs = -1;
	numoutputs = -1;
	numrules = -1;
	andtype = MINAND;
	ortype = MAXOR;
	imptype = MINAND;
	aggtype = MAXOR;
	defzytype = CENTROID;
	systype = MAMDANI;
	if (fuzzyinputs != NULL){
		delete(fuzzyinputs);
	}
	if (fuzzyoutputs != NULL){
		delete(fuzzyinputs);
	}
	if (rules != NULL){
		delete(fuzzyinputs);
	}

	int tmp;
	char tmpbuf1[STR_LEN], tmpbuf2[STR_LEN], tmpbuf3[STR_LEN], tmpbuf4[STR_LEN];
	while(AnalyzeNextEntry(fp)){
		if (Islabel) break;
		if (strncmp(namebuf, "Name", 4) == 0){
			if (getString()){
				strcpy(name, stringbuf);
			}
		}
		else if (strncmp(namebuf, "Type", 4) == 0){
			if (getString()){
				if (strncmp(stringbuf, "sugeno", 6) == 0){
					systype = SUGENO;
				}
				else if (strncmp(stringbuf, "mamdani", 7) == 0){
					systype = MAMDANI;
				}
				else{
					//printf("Illegal system type\n");
				}
			}
		}
		else if (strncmp(namebuf, "Version", 7) == 0){
		}
		else if (strncmp(namebuf, "NumInputs", 9) == 0){
			tmp = sscanf(valuebuf, "%d", &numinputs);
			if (tmp != 1 || numinputs < 0)
				return False;
		}
		else if (strncmp(namebuf, "NumOutputs", 10) == 0){
			tmp = sscanf(valuebuf, "%d", &numoutputs);
			if (tmp != 1 || numoutputs < 0)
				return False;
		}
		else if (strncmp(namebuf, "NumRules", 8) == 0){
			tmp = sscanf(valuebuf, "%d", &numrules);
			if (tmp != 1 || numrules < 0)
				return False;
		}
		else if (strncmp(namebuf, "AndMethod", 9) == 0){
			if (getString()){
				if (strncmp(stringbuf, "min", 3) == 0){
					andtype = MINAND;
				}
				else if (strncmp(stringbuf, "prod", 4) == 0){
					andtype = PRODAND;
				}
				else{
					//printf("Illegal andmethod\n");
				}
			}
		}
		else if (strncmp(namebuf, "OrMethod", 8) == 0){
			if (getString()){
				if (strncmp(stringbuf, "max", 3) == 0){
					ortype = MAXOR;
				}
				else if (strncmp(stringbuf, "probor", 6) == 0){
					ortype = PROBOR;
				}
				else{
					//printf("Illegal ormethod\n");
				}
			}
		}
		else if (strncmp(namebuf, "ImpMethod", 9) == 0){
			if (getString()){
				if (strncmp(stringbuf, "min", 3) == 0){
					imptype = MINAND;
				}
				else if (strncmp(stringbuf, "prod", 4) == 0){
					imptype = PRODAND;
				}
				else{
					//printf("Illegal impmethod\n");
				}
			}
		}
		else if (strncmp(namebuf, "AggMethod", 9) == 0){
			if (getString()){
				if (strncmp(stringbuf, "max", 3) == 0){
					aggtype = MAXOR;
				}
				else if (strncmp(stringbuf, "probor", 6) == 0){
					aggtype = PROBOR;
				}
				else{
					//printf("Illegal aggmethod\n");
				}
			}
		}
		else if (strncmp(namebuf, "DefuzzMethod", 12) == 0){
			if (getString()){
				if (strncmp(stringbuf, "centroid", 8) == 0){
					defzytype = CENTROID;
				}
				else if (strncmp(stringbuf, "wtsum", 5) == 0){
					defzytype = WEIGHTSUM;
				}
				else{
					//printf("Illegal defzytype\n");
				}
			}
		}
		else{
			//printf("Illegal inputs\n");
		}
	}
	if (numinputs <= 0 || numoutputs <= 0 || numrules <= 0) return False;

	int mfsidx;
/*	Read inputs	*/
	fuzzyinputs = new FuzzyVariable[numinputs];
	int i;
	for(i = 0; i < numinputs ; i++){
		while(AnalyzeNextEntry(fp)){
			if (Islabel) break;
			if (strncmp(namebuf, "Name", 4) == 0){
				if (getString()){
					strcpy(fuzzyinputs[i].name, stringbuf);
				}
			}
			else if (strncmp(namebuf, "Range", 5) == 0){
				tmp = sscanf(valuebuf, "[%f%f]", &fuzzyinputs[i].input_inf, &fuzzyinputs[i].input_sup);
				if (tmp != 2 || fuzzyinputs[i].input_inf > fuzzyinputs[i].input_sup)
					goto ExitReadInput;
			}
			else if (strncmp(namebuf, "NumMFs", 6) == 0){
				tmp = sscanf(valuebuf, "%d", &fuzzyinputs[i].nummfs);
				if (tmp != 1 || fuzzyinputs[i].nummfs <= 0)
					goto ExitReadInput;
				fuzzyinputs[i].mfs = new MemFunc[fuzzyinputs[i].nummfs];
				mfsidx = 0;
			}
			else{
			/*	Membership functions	*/
				if (fuzzyinputs[i].mfs == NULL){
					goto ExitReadInput;
				}
				tmp = sscanf(valuebuf, "'%[^']' %[^']' %[^']' %[^[] [ %[^]]", fuzzyinputs[i].mfs[mfsidx].name, tmpbuf1,
					tmpbuf2, tmpbuf3, tmpbuf4);
				if (tmp != 5)
					goto ExitReadInput;
				tmp = sscanf(tmpbuf4, "%f%f%f%f%f", &fuzzyinputs[i].mfs[mfsidx].params[0], 
						&fuzzyinputs[i].mfs[mfsidx].params[1], 
						&fuzzyinputs[i].mfs[mfsidx].params[2], 
						&fuzzyinputs[i].mfs[mfsidx].params[3], 
						&fuzzyinputs[i].mfs[mfsidx].params[4]);
				if (strncmp(tmpbuf2, "trimf", 5) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(TRIMF);
				}
				else if (strncmp(tmpbuf2, "gaussmf", 7) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(GAUSSMF);
				}
				else if (strncmp(tmpbuf2, "gauss2mf", 8) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(GAUSS2MF);
				}
				else if (strncmp(tmpbuf2, "trapmf", 6) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(TRAPMF);
				}
				else if (strncmp(tmpbuf2, "gbellmf", 7) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(GBELLMF);
				}
				else if (strncmp(tmpbuf2, "sigmf", 5) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(SIGMF);
				}
				else if (strncmp(tmpbuf2, "dsigmf", 6) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(DSIGMF);
				}
				else if (strncmp(tmpbuf2, "psigmf", 6) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(PSIGMF);
				}
				else if (strncmp(tmpbuf2, "zmf", 3) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(ZMF);
				}
				else if (strncmp(tmpbuf2, "pimf", 4) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(PIMF);
				}
				else if (strncmp(tmpbuf2, "smf", 3) == 0){
					fuzzyinputs[i].mfs[mfsidx].SetMFType(SMF);
				}
				else{
					//printf("Illegal membership function type\n");
				}
				if (tmp != fuzzyinputs[i].mfs[mfsidx].numparams)
					goto ExitReadInput;
				mfsidx ++;
			}
		}
		if (fuzzyinputs[i].nummfs <= 0 || fuzzyinputs[i].mfs == NULL)
			goto ExitReadInput;
	}
ExitReadInput:
	if (i != numinputs){
		delete fuzzyinputs;
		return False;
	}
/* Read Output	*/
	fuzzyoutputs = new FuzzyVariable[numoutputs];
	for(i = 0; i < numoutputs ; i++){
		if (!AnalyzeNextEntry(fp))
			goto ExitReadOutput;
		if (Islabel)
			continue;
		while(AnalyzeNextEntry(fp)){
			if (Islabel) break;
			if (strncmp(namebuf, "Name", 4) == 0){
				if (getString()){
					strcpy(fuzzyoutputs[i].name, stringbuf);
				}
			}
			else if (strncmp(namebuf, "Range", 5) == 0){
				tmp = sscanf(valuebuf, "[%f%f]", &fuzzyoutputs[i].input_inf, &fuzzyoutputs[i].input_sup);
				if (tmp != 2 || fuzzyoutputs[i].input_inf > fuzzyoutputs[i].input_sup)
					goto ExitReadOutput;
			}
			else if (strncmp(namebuf, "NumMFs", 6) == 0){
				tmp = sscanf(valuebuf, "%d", &fuzzyoutputs[i].nummfs);
				if (tmp != 1 || fuzzyoutputs[i].nummfs <= 0)
					goto ExitReadOutput;
				fuzzyoutputs[i].mfs = new MemFunc[fuzzyoutputs[i].nummfs];
				mfsidx = 0;
			}
			else{
			/*	Membership functions	*/
				if (fuzzyoutputs[i].mfs == NULL){
					goto ExitReadOutput;
				}
				tmp = sscanf(valuebuf, "'%[^']' %[^']' %[^']' %[^[] [ %[^]]", fuzzyoutputs[i].mfs[mfsidx].name, tmpbuf1,
					tmpbuf2, tmpbuf3, tmpbuf4);
				if (tmp != 5)
					goto ExitReadOutput;
				tmp = sscanf(tmpbuf4, "%f%f%f%f%f", &fuzzyoutputs[i].mfs[mfsidx].params[0], 
						&fuzzyoutputs[i].mfs[mfsidx].params[1], 
						&fuzzyoutputs[i].mfs[mfsidx].params[2], 
						&fuzzyoutputs[i].mfs[mfsidx].params[3], 
						&fuzzyoutputs[i].mfs[mfsidx].params[4]);

				if (strncmp(tmpbuf2, "trimf", 5) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(TRIMF);
				}
				else if (strncmp(tmpbuf2, "gaussmf", 7) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(GAUSSMF);
				}
				else if (strncmp(tmpbuf2, "gauss2mf", 8) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(GAUSS2MF);
				}
				else if (strncmp(tmpbuf2, "trapmf", 6) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(TRAPMF);
				}
				else if (strncmp(tmpbuf2, "gbellmf", 7) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(GBELLMF);
				}
				else if (strncmp(tmpbuf2, "sigmf", 5) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(SIGMF);
				}
				else if (strncmp(tmpbuf2, "dsigmf", 6) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(DSIGMF);
				}
				else if (strncmp(tmpbuf2, "psigmf", 6) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(PSIGMF);
				}
				else if (strncmp(tmpbuf2, "zmf", 3) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(ZMF);
				}
				else if (strncmp(tmpbuf2, "pimf", 4) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(PIMF);
				}
				else if (strncmp(tmpbuf2, "smf", 3) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(SMF);
				}
				else if (strncmp(tmpbuf2, "constant", 8) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(CONSTANT);
				}
				else if (strncmp(tmpbuf2, "linear", 8) == 0){
					fuzzyoutputs[i].mfs[mfsidx].SetMFType(LINEAR);
				}
				else{
					//printf("Illegal membership function type\n");
				}
				if (tmp != fuzzyoutputs[i].mfs[mfsidx].numparams)
					goto ExitReadOutput;
				mfsidx ++;
			}
		}
		if (fuzzyoutputs[i].nummfs <= 0 || fuzzyoutputs[i].mfs == NULL)
			goto ExitReadOutput;
	}
ExitReadOutput:
	if (i != numoutputs){
		delete fuzzyoutputs;
		return False;
	}
/*	Read rules	*/
	rules = new Rule[numrules];
	for(i = 0; i < numrules ; i++){
		rules[i].SetRuleParams(numinputs, numoutputs);
		if (!AnalyzeNextEntry(fp))
			goto ExitReadRule;
		if (Islabel) continue;
		int p_buf = 0;
		int j;
		for(j = 0; j < numinputs; j ++){
			if (sscanf(buf + p_buf, "%d", &rules[i].inputmfidx[j]) != 1)
				goto ExitReadRule;
			while( (*(buf + p_buf) < '0' || *(buf + p_buf) > '9') && *(buf + p_buf) !='-' && *(buf + p_buf) !=0 )
				p_buf ++;
			while( *(buf + p_buf) >= '0' && *(buf + p_buf) <= '9')
				p_buf ++;
		}

		while( *(buf + p_buf) != ',' && *(buf + p_buf) != 0)
			p_buf ++;
		p_buf ++;
		for(j = 0; j < numoutputs; j ++){
			if (sscanf(buf + p_buf, "%d", &rules[i].outputmfidx[j]) != 1)
				goto ExitReadRule;
			while( (*(buf + p_buf) < '0' || *(buf + p_buf) > '9') && *(buf + p_buf) !='-' && *(buf + p_buf) !=0 )
				p_buf ++;
			while( *(buf + p_buf) >= '0' && *(buf + p_buf) <= '9')
				p_buf ++;
		}
		if (sscanf(buf + p_buf, "%[^(](%f", tmpbuf1, &rules[i].weight) != 2)
			goto ExitReadRule;
		if (sscanf(buf + p_buf, "%[^:]:%d", tmpbuf1, &tmp) != 2)
			goto ExitReadRule;
		if (tmp == 1)
			rules[i].rc_type = ANDRULE;
		else if (tmp == 2)
			rules[i].rc_type = ORRULE;
		else{
			//printf("Illegal rctype\n");
		}

	}
ExitReadRule:
	if (i != numrules){
		delete rules;
		return False;
	}
	return True;
}

float FIS::AND(float x, float y){
	return AND(x, y, andtype);
}

float FIS::AND(float x, float y, AndMethod and_type){
	switch(and_type){
	case MINAND:	return x < y ? x : y;
	case PRODAND:	return x * y;
	}
	return x < y ? x : y;		
}

float FIS::OR(float x, float y){
	return OR(x, y, ortype);
}

float FIS::OR(float x, float y, ORMethod or_type){
	switch(or_type){
	case MAXOR:		return x > y ? x : y;
	case PROBOR:	return x + y - x * y;
	}
	return x > y ? x : y;
}

void FIS::SetInput(int idx, float input){
	if (idx <= 0 || idx > numinputs)
		return;
	fuzzyinputs[idx - 1].SetInput(input);
}

void FIS::SetInput(float* input){
	int idx;
	for(idx = 1; idx <= numinputs; idx ++){
		fuzzyinputs[idx -1].SetInput(input[idx -1]);
	}
}

void FIS::Fuzzify(){
	int idx;
	for(idx = 1; idx <= numinputs; idx ++){
		fuzzyinputs[idx -1].Fuzzify();
	}
}

void FIS::ApplyRule(int idx){
	if (idx <= 0 || idx > numrules)
		return;
	float fire, tmp;
	if (rules[idx -1].rc_type == ANDRULE){
		fire = 1;
		int i;
		for(i = 1; i <= numinputs; i ++){
			if (rules[idx -1].inputmfidx[i -1] == 0) continue;
			tmp = fuzzyinputs[i -1].GetOutput(abs(rules[idx -1].inputmfidx[i -1]));
			if (rules[idx -1].inputmfidx[i -1] < 0)
				tmp = NOT(tmp);
			fire = AND(fire, tmp);
		}
	}
	else{
		fire = 0;
		int i;
		for(i = 1; i <= numinputs; i ++){
			if (rules[idx -1].inputmfidx[i -1] == 0) continue;
			tmp = fuzzyinputs[i -1].GetOutput(abs(rules[idx -1].inputmfidx[i -1]));
			if (rules[idx -1].inputmfidx[i -1] < 0)
				tmp = NOT(tmp);
			fire = OR(fire, tmp);
		}
	}
	rules[idx -1].firestrength = fire;
}

float FIS::Imp_Defuzzy(int idx){
	if (idx <= 0 || idx > numoutputs)
		return 0.0f;
	switch(defzytype){
	case CENTROID:
		return CentroidDefzy(idx);
		break;
	}
	return 0.0f;
}

/*	Intergral Simpson	*/
//no is the indicator of the different weight functions
float FIS::Simpson(int n, float a, float b, int no){
	if(n <= 0) return 0.0f;
	float h = (float) (b - a)/n, x;
	float result = WeightFunc(b, no)+WeightFunc(a, no);

	x = h/2 + a;
	int k;
	for(k = 0; k < n; k++)
		result += 4 * WeightFunc(x, no), x += h;

	x= h + a;
	for(k = 1; k < n; k++)
		result += 2 * WeightFunc(x, no), x += h;

	result *= h/6;
	return result;
}

float FIS::Gauss(int n, float a, float b, int no){
	if (n<=0) return 0;
	float h = (b-a) /n;
	float result = 0;
	int i;
	for(i=0; i < n; i++){
		result += WeightFunc((i+0.5f)*h+a-h/(2.0f*(float)sqrt(3.0f)), no)+WeightFunc((i+0.5f)*h+a+h/(2.0f*(float)sqrt(3.0f)), no);
	}
	result *= h/2;
	return result;
}

float FIS::Romberg(int n, float a, float b, int no){
	if (n<=0) return 0;
	float* T = new float[n];
	float h = b - a;
	int i,  k, t;
	T[0] = h*(WeightFunc(a, no) + WeightFunc(b, no))/2;
	t = 1;
	for(k=1; k < n; k++){
		T[k] = T[k-1]/2;
		for(i=0; i<t; i++){
			T[k] += h/2*WeightFunc(a+h*(i+0.5f), no);
		}
		h /= 2;
		t *= 2;
	}
	t = 4;
	for (k = 1; k < n; k++){
		for(i = 0; i< n - k; i++)
			T[i] = (t*T[i+1] - T[i]) / (t-1);
		t *= 4;
	}
	float result = T[0];
	delete T;
	return result;
}

/*	WeightFunc	*/
float FIS::WeightFunc(float x, int no){
	float weight, tmp;
	weight = 0.0f;
	int i;
	for(i = 1; i <= numrules; i ++){
		if (rules[i -1].outputmfidx[outputidx -1] == 0) continue;
		tmp = fuzzyoutputs[outputidx -1].GetMembership(abs(rules[i -1].outputmfidx[outputidx -1]), x);
		if (rules[i -1].outputmfidx[outputidx -1] < 0)
			tmp = NOT(tmp);
		weight = OR(weight, AND(rules[i -1].GetFirestrength(), tmp, imptype), aggtype);
	}
	switch(no){
	case 0:
		break;
	case 1:
		weight *= x;
		break;
	}
	return weight;
}

float FIS::CentroidDefzy(int idx){	
	outputidx = idx;
	float sumweight;
	
	sumweight = Simpson(10, 0, 1, 0);
	if (sumweight == 0) return 0.0f;
	return Simpson(10, 0, 1, 1)/ sumweight;
}

void FIS::FISRun(float *input, float * output){
	SetInput(input);
	Fuzzify();
	int idx;
	for(idx = 1; idx <= numrules; idx ++){
		ApplyRule(idx);
	}

	for(idx = 1; idx <= numoutputs; idx ++){
		output[idx -1] = Imp_Defuzzy(idx);
	}
}

